/**
 * Copyright (c) 2023 Dominic Masters
 * 
 * This software is released under the MIT License.
 * https://opensource.org/licenses/MIT
 */

#include "shader.h"
#include "assert/assert.h"
#include "assert/assertgl.h"

int32_t SHADER_ACTIVE_COUNT;

void shaderInit(shader_t *shader) {
  assertNotNull(shader, "shaderInit: shader cannot be null.");

  shader->shaderVertex = -1;
  shader->shaderFrag = -1;
  shader->shaderProgram = -1;

  SHADER_ACTIVE_COUNT++;
}

void shaderCompile(
  shader_t *shader,
  const char_t *vertex,
  const char_t *fragment
) {
  assertNotNull(shader, "shaderCompile: shader cannot be null.");
  assertStringValid(vertex, 2048, "shaderGetParameterByName: name invalid.");
  assertStringValid(fragment, 2048, "shaderGetParameterByName: name invalid.");
  assertTrue(shader->shaderVertex == -1, "shaderCompile: vertex not -1.");
  assertTrue(shader->shaderFrag == -1, "shaderCompile: fragment not -1.");
  assertTrue(shader->shaderProgram == -1, "shaderCompile: program not -1.");

  GLint isSuccess;
  int32_t maxLength;
  char_t error[2048];
  
  // Load the vertex shader first
  shader->shaderVertex = glCreateShader(GL_VERTEX_SHADER);
  assertNoGLError();
  glShaderSource(shader->shaderVertex, 1, &vertex, NULL);
  assertNoGLError();
  glCompileShader(shader->shaderVertex);
  assertNoGLError();

  // Validate
  glGetShaderiv(shader->shaderVertex, GL_COMPILE_STATUS, &isSuccess);
  assertNoGLError();
  if(!isSuccess) {
    glGetShaderiv(shader->shaderVertex, GL_INFO_LOG_LENGTH, &maxLength);
    glGetShaderInfoLog(shader->shaderVertex, maxLength, &maxLength, error);
    sprintf(
      error,
      "shaderCompile: Failed to compile vert shader:\n%s",
      error
    );
    assertUnreachable(error);
  }
  assertNoGLError();

  // Now load the Frag shader
  shader->shaderFrag = glCreateShader(GL_FRAGMENT_SHADER);
  assertNoGLError();
  glShaderSource(shader->shaderFrag, 1, &fragment, NULL);
  assertNoGLError();
  glCompileShader(shader->shaderFrag);
  assertNoGLError();

  glGetShaderiv(shader->shaderFrag, GL_COMPILE_STATUS, &isSuccess);
  assertNoGLError();
  if(!isSuccess) {
    glGetShaderiv(shader->shaderFrag, GL_INFO_LOG_LENGTH, &maxLength);
    glGetShaderInfoLog(shader->shaderFrag, maxLength, &maxLength, error);
    glDeleteShader(shader->shaderVertex);
    sprintf(
      error,
      "shaderCompile: Failed to compile frag shader:\n%s",
      error
    );
    assertUnreachable(error);
  }
  assertNoGLError();

  // Now create the shader program.
  shader->shaderProgram = glCreateProgram();
  assertNoGLError();
  glAttachShader(shader->shaderProgram, shader->shaderVertex);
  assertNoGLError();
  glAttachShader(shader->shaderProgram, shader->shaderFrag);
  assertNoGLError();

  //Bind, Verify & Use the shader program
  glLinkProgram(shader->shaderProgram);
  assertNoGLError();

  glGetProgramiv(shader->shaderProgram, GL_LINK_STATUS, &isSuccess);
  assertNoGLError();
  if(!isSuccess) {
    glGetProgramiv(shader->shaderProgram, GL_INFO_LOG_LENGTH, &maxLength);
    assertNoGLError();
    glGetProgramInfoLog(shader->shaderProgram, maxLength, &maxLength, error);
    assertNoGLError();
    glDeleteShader(shader->shaderVertex);
    assertNoGLError();
    glDeleteShader(shader->shaderFrag);
    assertNoGLError();
    sprintf(
      error,
      "shaderCompile: Error compiling shader program:\n%s",
      error
    );
    assertUnreachable(error);
  }
}

shaderparameter_t shaderGetParameterByName(
  const shader_t *shader,
  const char_t *name
) {
  assertNotNull(shader, "shaderGetParameterByName: shader cannot be null.");
  assertStringValid(name, 128, "shaderGetParameterByName: name invalid.");

  shaderparameter_t param = glGetUniformLocation(shader->shaderProgram, name);
  assertNoGLError();
  assertTrue(param != -1, "shaderGetParameterByName: param not found.");
  return param;
}

void shaderBind(const shader_t *shader) {
  assertNotNull(shader, "shaderBind: Shader cannot be null.");
  assertTrue(
    shader->shaderProgram != -1,
    "shaderBind: Cannot bind a program that is not ready"
  );
  glUseProgram(shader->shaderProgram);
  assertNoGLError();
}

void shaderSetMatrix(const shaderparameter_t param, const mat4_t matrix) {
  glUniformMatrix4fv(param, 1, GL_FALSE, matrix[0]);
  assertNoGLError();
}

void shaderSetBoolean(const shaderparameter_t param, const bool_t value) {
  glUniform1i(param, value);
  assertNoGLError();
}

void shaderSetInteger(const shaderparameter_t param, const int32_t value) {
  glUniform1i(param, value);
  assertNoGLError();
}

void shaderSetFloat(const shaderparameter_t param, const float_t value) {
  glUniform1f(param, value);
  assertNoGLError();
}

void shaderSetVector2(const shaderparameter_t param, const vec2_t vector) {
  glUniform2f(param, vector[0], vector[1]);
  assertNoGLError();
}

void shaderSetVector3(const shaderparameter_t param, const vec3_t vector) {
  glUniform3f(param, vector[0], vector[1], vector[2]);
  assertNoGLError();
}

void shaderSetVector4(const shaderparameter_t param, const vec4_t vector) {
  glUniform4f(param, vector[0], vector[1], vector[2], vector[3]);
  assertNoGLError();
}

void shaderSetColor3f(const shaderparameter_t param, const color3f_t color) {
  glUniform3f(param, color[0], color[1], color[2]);
  assertNoGLError();
}

void shaderSetColor4f(const shaderparameter_t param, const color4f_t color) {
  glUniform4f(param, color[0], color[1], color[2], color[3]);
  assertNoGLError();
}

void shaderSetTexture(const shaderparameter_t param, const textureslot_t slot) {
  glUniform1i(param, slot);
  assertNoGLError();
}

void shaderDispose(const shader_t *shader) {
  if(shader->shaderProgram != -1) {
    glDeleteProgram(shader->shaderProgram);
    assertNoGLError();
  }

  if(shader->shaderVertex != -1) {
    glDeleteShader(shader->shaderVertex);
    assertNoGLError();
  }

  if(shader->shaderFrag != -1) {
    glDeleteShader(shader->shaderFrag);
    assertNoGLError();
  }

  SHADER_ACTIVE_COUNT--;
}