/**
 * Copyright (c) 2023 Dominic Masters
 * 
 * This software is released under the MIT License.
 * https://opensource.org/licenses/MIT
 */

#pragma once
#include "dawn.h"
#include "dawnopengl.h"
#include "display/color.h"
#include "display/texture.h"

typedef GLuint shaderparameter_t;
extern int32_t SHADER_ACTIVE_COUNT;

typedef struct {
  /** Pointer to an uploaded vertex shader program */
  GLuint shaderVertex;

  /** Pointer to an uploaded fragment shader program */
  GLuint shaderFrag;

  /** Pointer to an uploaded shader program linked */
  GLuint shaderProgram;
} shader_t;

/**
 * Initializes a shader object.
 * 
 * @param shader The shader object to initialize.
 */
void shaderInit(shader_t *shader);

/**
 * Compiles a shader object.
 * 
 * @param shader The shader object to compile.
 * @param vertex The vertex shader source.
 * @param fragment The fragment shader source.
 */
void shaderCompile(
  shader_t *shader,
  const char_t *vertex,
  const char_t *fragment
);

/**
 * Returns a shader parameter by name.
 * 
 * @param shader Shader to get the parameter from.
 * @param name Name of the parameter to get.
 * @return The shader parameter.
 */
shaderparameter_t shaderGetParameterByName(
  const shader_t *shader,
  const char_t *name
);

/**
 * Attaches the supplied shader as the current shader.
 * 
 * @param shader 
 */
void shaderBind(const shader_t *shader);

/**
 * Sets a matrix parameter on the shader.
 * 
 * @param param Parameter to set.
 * @param matrix Matrix value to set.
 */
void shaderSetMatrix(const shaderparameter_t param, const mat4_t matrix);

/**
 * Sets a boolean parameter on the shader.
 * 
 * @param param Parameter to set.
 * @param value Boolean value to set.
 */
void shaderSetBoolean(const shaderparameter_t param, const bool_t value);

/**
 * Sets an integer parameter on the shader.
 * 
 * @param param Parameter to set.
 * @param value Integer value to set.
 */
void shaderSetInteger(const shaderparameter_t param, const int32_t value);

/**
 * Sets a float parameter on the shader.
 * 
 * @param param Parameter to set.
 * @param value Float value to set.
 */
void shaderSetFloat(const shaderparameter_t param, const float_t value);

/**
 * Sets a vector parameter on the shader.
 * 
 * @param param Parameter to set.
 * @param vector Vector value to set.
 */
void shaderSetVector2(const shaderparameter_t param, const vec2_t vector);

/**
 * Sets a vector parameter on the shader.
 * 
 * @param param Parameter to set.
 * @param vector Vector value to set.
 */
void shaderSetVector3(const shaderparameter_t param, const vec3_t vector);

/**
 * Sets a vector parameter on the shader.
 * 
 * @param param Parameter to set.
 * @param vector Vector value to set.
 */
void shaderSetVector4(const shaderparameter_t param, const vec4_t vector);

/**
 * Sets a color parameter on the shader.
 * 
 * @param param Parameter to set.
 * @param color Color value to set.
 */
void shaderSetColor3f(const shaderparameter_t param, const color3f_t color);

/**
 * Sets a color parameter on the shader.
 * 
 * @param param Parameter to set.
 * @param color Color value to set.
 */
void shaderSetColor4f(const shaderparameter_t param, const color4f_t color);

/**
 * Sets a texture parameter on the shader.
 * 
 * @param param Parameter to set.
 * @param slot Slot to bind the texture to.
 */
void shaderSetTexture(const shaderparameter_t param, const textureslot_t slot);

/**
 * Disposes of a shader object.
 * 
 * @param shader Shader to destroy.
 */
void shaderDispose(const shader_t *shader);