/**
 * Copyright (c) 2023 Dominic Masters
 * 
 * This software is released under the MIT License.
 * https://opensource.org/licenses/MIT
 */

#include "texture.h"
#include "assert/assert.h"
#include "assert/assertgl.h"
#include "util/math.h"

int32_t TEXTURE_ACTIVE_COUNT;

void textureInit(
  texture_t *texture,
  int32_t width,
  int32_t height,
  textureformat_t format,
  texturedataformat_t dataFormat
) {
  assertNotNull(texture, "textureInit: texture cannot be null.");
  assertTrue(width > 0, "textureInit: width must be greater than 0.");
  assertTrue(height > 0, "textureInit: height must be greater than 0.");

  texture->width = width;
  texture->height = height;
  texture->format = format;
  texture->dataFormat = dataFormat;
  texture->wrapModeX = TEXTURE_WRAP_MODE_REPEAT;
  texture->wrapModeY = TEXTURE_WRAP_MODE_REPEAT;
  texture->filterModeMin = TEXTURE_FILTER_MODE_LINEAR;
  texture->filterModeMag = TEXTURE_FILTER_MODE_LINEAR;
  
  int32_t maxSize;
  glGetIntegerv(GL_MAX_TEXTURE_SIZE, &maxSize);
  assertTrue(
    width > 0 && width <= maxSize, "textureInit: Width is out of bounds!"
  );
  assertTrue(
    height > 0 && height <= maxSize, "textureInit: Height is out of bounds!"
  );

  glGenTextures(1, &texture->id);
  assertNoGLError();
  assertTrue(texture->id > 0, "textureInit: Failed to generate texture ID!");

  TEXTURE_ACTIVE_COUNT++;

  switch(texture->dataFormat) {
    case TEXTURE_DATA_FORMAT_FLOAT:
      texture->elementSize = sizeof(float_t);
      break;
    case TEXTURE_DATA_FORMAT_UNSIGNED_BYTE:
      texture->elementSize = sizeof(uint8_t);
      break;
    default:
      assertUnreachable("textureInit: Unknown data format.");
      break;
  }
  switch(texture->format) {
    case TEXTURE_FORMAT_RGBA:
      texture->elementSize *= 4;
      break;
    case TEXTURE_FORMAT_RGB:
      texture->elementSize *= 3;
      break;
    case TEXTURE_FORMAT_RG:
      texture->elementSize *= 2;
      break;
    case TEXTURE_FORMAT_R:
      break;
    default:
      assertUnreachable("textureInit: Unknown texture format.");
      break;
  }
  
  textureBind(texture, 0);
  textureBuffer(texture, NULL);
}

void textureBind(const texture_t *texture, const textureslot_t slot) {
  glActiveTexture(GL_TEXTURE0 + slot);
  assertNoGLError();
  glBindTexture(GL_TEXTURE_2D, texture->id);
  assertNoGLError();
  
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, texture->wrapModeX);
  assertNoGLError();
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, texture->wrapModeY);
  assertNoGLError();

  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, texture->filterModeMin);
  assertNoGLError();
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, texture->filterModeMag);
  assertNoGLError();
}

void textureBuffer(const texture_t *texture, const void *data) {
  assertNotNull(texture, "textureBuffer: texture cannot be null.");

  textureBind(texture, 0);
  glTexImage2D(
    GL_TEXTURE_2D,
    0,
    texture->format,
    texture->width,
    texture->height,
    0,
    texture->format,
    texture->dataFormat,
    data
  );
  assertNoGLError();
  
  // glGenerateMipmap(GL_TEXTURE_2D);
  // assertNoGLError();
}

void textureBufferSub(
  const texture_t *texture,
  const void *data,
  const size_t offset,
  const size_t length
) {
  assertNotNull(texture, "textureBufferSub: Texture cannot be null.");
  assertNotNull(data, "textureBufferSub: Data cannot be null.");
  assertTrue(offset >= 0, "textureBufferSub: Offset must be greater than 0.");
  assertTrue(length > 0, "textureBufferSub: Length must be greater than 0.");
  assertTrue(
    offset + length <= texture->width * texture->height,
    "textureBufferSub: Buffer out of bounds."
  );

  if(offset == 0 && length == texture->width * texture->height) {
    textureBuffer(texture, data);
    return;
  }

  // Bind the texture.
  textureBind(texture, 0);

  // Buffer initial row only.
  int32_t xOffset = offset % texture->width;
  int32_t yOffset = offset / texture->width;
  size_t n = mathMin(length, texture->width - xOffset);
  size_t buffered = 0;
  glTexSubImage2D(
    GL_TEXTURE_2D,
    0,
    xOffset, yOffset,
    n, 1,
    texture->format, texture->dataFormat,
    data
  );
  assertNoGLError();
  buffered += n;
  yOffset += 1;

  // Buffer the creamy middle (inefficient?)
  int32_t rows = 0;
  size_t initialBuffered = buffered;
  while(buffered + texture->width < length) {
    rows += 1;
    buffered += texture->width;
  }

  if(rows > 0) {
    glTexSubImage2D(
      GL_TEXTURE_2D,
      0,
      0, yOffset,
      texture->width, rows,
      texture->format, texture->dataFormat,
      data + (initialBuffered * texture->elementSize)
    );
    assertNoGLError();
    yOffset += rows;
  }

  // Buffer last row only.
  if(buffered != length) {
    glTexSubImage2D(
      GL_TEXTURE_2D,
      0,
      0, yOffset,
      length - buffered, 1,
      texture->format, texture->dataFormat,
      data + (buffered * texture->elementSize)
    );
    assertNoGLError();
  }
}

void textureDispose(const texture_t *texture) {
  glDeleteTextures(1, &texture->id);
  assertNoGLError();
  TEXTURE_ACTIVE_COUNT--;
}