/**
 * Copyright (c) 2024 Dominic Masters
 * 
 * This software is released under the MIT License.
 * https://opensource.org/licenses/MIT
 */

#include "textureframebuffer.h"
#include "assert/assert.h"
#include "assert/assertgl.h"

void textureFrameBufferInit(
  textureframebuffer_t *framebuffer,
  const int32_t width,
  const int32_t height
) {
  assertNotNull(framebuffer, "Framebuffer cannot be NULL");
  assertTrue(width > 0, "Width must be greater than 0");
  assertTrue(height > 0, "Height must be greater than 0");

  textureInit(
    &framebuffer->texture,
    width, height,
    TEXTURE_FORMAT_RGBA,
    TEXTURE_DATA_FORMAT_FLOAT
  );

  // Create Frame Buffer
  glGenFramebuffers(1, &framebuffer->fboId);
  assertNoGLError();
  glBindFramebuffer(GL_FRAMEBUFFER, framebuffer->fboId);
  assertNoGLError();
  glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0,
    GL_TEXTURE_2D, framebuffer->texture.id, 0
  );
  assertNoGLError();

  // Create Render Buffer
  glGenRenderbuffers(1, &framebuffer->rboId);
  assertNoGLError();
  glBindRenderbuffer(GL_RENDERBUFFER, framebuffer->rboId); 
  assertNoGLError();
  glRenderbufferStorage(
    GL_RENDERBUFFER, GL_DEPTH24_STENCIL8,
    framebuffer->texture.width, framebuffer->texture.height
  );
  assertNoGLError();
  glBindRenderbuffer(GL_RENDERBUFFER, 0);
  assertNoGLError();
  glFramebufferRenderbuffer(GL_FRAMEBUFFER, GL_DEPTH_STENCIL_ATTACHMENT,
    GL_RENDERBUFFER, framebuffer->rboId
  );
  assertNoGLError();

  // Validate things went correct.
  if(glCheckFramebufferStatus(GL_FRAMEBUFFER) != GL_FRAMEBUFFER_COMPLETE) {
    assertUnreachable("Framebuffer is not complete!");
  }
}

void textureFrameBufferBind(const textureframebuffer_t *framebuffer) {
  assertNotNull(framebuffer, "Framebuffer cannot be NULL");

  glBindFramebuffer(GL_FRAMEBUFFER, framebuffer->fboId);
  assertNoGLError();
  glViewport(0, 0, framebuffer->texture.width, framebuffer->texture.height);
  assertNoGLError();
}

void textureFrameBufferClear(const textureframebuffer_t *framebuffer) {
  assertNotNull(framebuffer, "Framebuffer cannot be NULL");

  glClearColor(0,0,0,1);
  assertNoGLError();
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  assertNoGLError();
}

void textureFrameBufferDispose(textureframebuffer_t *framebuffer) {
  assertNotNull(framebuffer, "Framebuffer cannot be NULL");
  
  glDeleteFramebuffers(1, &framebuffer->fboId);
  glDeleteRenderbuffers(1, &framebuffer->rboId);
  textureDispose(&framebuffer->texture);
}